// Module Name: maze.cpp
// Author:      Stephen Manley (smanley@nyx.net)
//
// Description:
// 
// This program generates a maze from the data files created with the 
// mazemaker. It takes the 2D maze information which is stored in 
// terms of an array of color values and generates a 3D maze from it.
// This 3D maze is comprised of walls, drawn with one of three methods.
// There is basic C++ class abstraction, but since it wasn't emphasised
// on the assignment, I didn't go crazy.
//
// The program loads the maze, computes what walls need to go where, 
// then stores the information in the maze for optimal redraw (no 
// additional computation.
// 
// Usage:
// 
// Pressing l will load the maze into the maze generator and automatically
// display the new 3D generated maze. The file must be stored in:
// "./grid.dat" This is also done automagically on startup.
// 
// To maniuplate the maze, q, w, a, s rotate in the X and Y parameters. 
// c and v scales the maze. Alternately, pressing and holding the mouse
// button and moving up/dowwn will tilt the maze. The maze automatically
// rotates about the axis. Moving the mouse in the horizontal direction
// will change the scale of the drawing.
// 
// My machine is quite slow, so the keyboard controls work much better
// than the mouse, although the mouse controls work fine (Just not
// realtime on my machine). 
// 
// NOTE ON MOUSE CONTROL: For optimal use, click the left button in the
// middle of the window. Then slowly move towards the edge of the screen.
// The response isn't exactly linear, but this approach made for the 
// most managable combination of scale and rotate on my slow machine.
//
////////////////////////////////////////////////////////////////////////////

// Eeeewwwww
#include <windows.h>

// System includes
#include <GL/gl.h>
#include <GL/glut.h>
#include <iostream.h>
#include <fstream.h>

// Projection and Window Information
double left   = 0.0;  
double right  = 30.0;
double top    = 0.0;
double bottom = 25.0;
double near1   = 1.0;
double far1    = 40.0;

// Height and width of the active window
int windowHeight, windowWidth;

// Default background color
double rBack = 0.0;
double gBack = 0.0;
double bBack = 0.0;
 
// Rotation, scale and translation information
double xRotation = 0.0;
double yRotation = 0.0;

double xScale = 1.2;
double yScale = 1.2;
double zScale = 1.2;

// Maze Size & Data
int xSize, ySize;
double* tileArray;

// Mouse status variables
int mouseClickX, mouseClickY;
int leftMouseDown = 0;

// Animation Control
int waiting = 0;

// Function Prototypes
void drawPlane(double z);

void drawWall1(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4);

void drawWall2(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4);

void drawWall3(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4);


// Configure projection information
void init(void) 
{
   glClearColor (0.0, 0.0, 0.0, 0.0);
   glShadeModel (GL_SMOOTH);

   // Initialize projection
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(45.0, 1.0, 1.0, 100.0);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

// Redraw the whole scene
void display(void)
{
   // Clear the old display
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // Apply model transformations
   glPushMatrix();

   glTranslated(0,0,-60);
   glRotated(xRotation, 1., 0., 0.);
   glRotated(yRotation, 0., 0., 1.);
   glScaled(xScale, yScale, zScale);

   // Draw the ceiling (0.01 to win roundoff with wall)
   drawPlane(0.01);

   int i = -1;
   int tilesize = 1;

   double z = 0; // Ceiling Plane
   double z2 = -5; // Floor Plane


   // Draw the walls - there are two walls for each ceiling tile.
   for (int x=-10; x<xSize-10; x++)
   {
       for (int y=-10; y<ySize-10; y++)
       {
          double r = tileArray[(++i)]; 
          double b = tileArray[(++i)];
          double g = tileArray[(++i)];

		  // Depending on what color was selected in the 2D maze creator,
		  // draw it with an appropriately styled tile.
          if (r == 1)
		  {              
			  // Top and bottom
   		      drawWall3(x,y,z, x,y,z2, x+1,y,z2, x+1,y,z );
   		      drawWall3(x,y-1,z, x,y-1,z2, x+1,y-1,z2, x+1,y-1,z );

              // Sides
   		      drawWall3(x+1,y,z, x+1,y,z2, x+1,y-1,z2, x+1,y-1,z);
   		      drawWall3(x,y,z, x,y,z2, x,y-1,z2, x,y-1,z);
   		  }
   		  
          if (g == 1)
		  {
			  // Top and bottom
   		      drawWall2(x,y,z, x,y,z2, x+1,y,z2, x+1,y,z );
   		      drawWall2(x,y-1,z, x,y-1,z2, x+1,y-1,z2, x+1,y-1,z );

              // Sides
   		      drawWall2(x+1,y,z, x+1,y,z2, x+1,y-1,z2, x+1,y-1,z);
   		      drawWall2(x,y,z, x,y,z2, x,y-1,z2, x,y-1,z);
   		  }
          if (b == 1)
		  {
			  // Top and bottom
   		      drawWall1(x,y,z, x,y,z2, x+1,y,z2, x+1,y,z );
   		      drawWall1(x,y-1,z, x,y-1,z2, x+1,y-1,z2, x+1,y-1,z );

              // Sides
   		      drawWall1(x+1,y,z, x+1,y,z2, x+1,y-1,z2, x+1,y-1,z);
   		      drawWall1(x,y,z, x,y,z2, x,y-1,z2, x,y-1,z);
   		  }

       }
   }

   // Draw the ceiling
   drawPlane(-5.01);

   glPopMatrix();

   // Swap and flush the buffers
   glutSwapBuffers();
}

// Reshape the window - save the parameters
void reshape (int w, int h)
{
   glViewport (0, 0, (GLsizei) w, (GLsizei) h);

   windowWidth = w;
   windowHeight = h; 
}

// Loads the maze from the files created by mazemaker into a 2D array
void loadMaze()
{
    fstream inFile("grid.dat", ios::in);
    
	if (!inFile)
		return;

    // Read in the grid dimensions
    inFile >> xSize;
    inFile >> ySize;
    tileArray = new double[xSize*ySize*3];

    double r,g,b;

	// Read in the colors of each tile (X*Y)
    for (int i=0; i<xSize*ySize*3; i++)
    {
        inFile >> r;
        tileArray[i] = r;
        
        inFile >> g;
		tileArray[(++i)] = g;

        inFile >> b;
		tileArray[(++i)] = b;
    }

    inFile.close();
}

// Keyboard handler
void keyboard(unsigned char key, int x, int y)
{
   switch (key) {
      case 27:
         exit(0);
         break;

      case 'l':
         loadMaze();
         display(); // redraw
         break;   

      case 'c':
         xScale += .1;
         yScale += .1;
         zScale += .1;
         
         display();
         break;

      case 'v':
         xScale -= .1;
         yScale -= .1;
         zScale -= .1;
         
         display();
         break;

      case 'q':
		 xRotation += 5;
      	 display();
      	 break;

      case 'w': 	 
		 xRotation -= 5;
     	 display();
      	 break; 

      case 'a':
         yRotation += 5;
         display();
         break;	 

      case 's':
         yRotation -= 5;
         display();
         break;	 
   }
}

// Mouse motion callback
void mouseMove(int x, int y)
{
	if (leftMouseDown)
	{
		// Calculate a delta X and delta Y
		double dx = mouseClickX - x;
		double dy = mouseClickY - y;

		// Allow a maximum scale and rotation
		double dScale =  abs((dx/((double)windowWidth) * 3.0));
		double dRotate = abs((dy/((double)windowWidth) * 20.0));

		//dScale = 0.2;
		//dRotate = 0.0;

		// Left right changes scale - put some careful bounds here
		if(mouseClickX < x) 
		{
			xScale += dScale;
			yScale += dScale;
			zScale += dScale;
		}

		if(mouseClickX > x)
		{
			xScale -= dScale;
			yScale -= dScale;
			zScale -= dScale;
		}

		// Up down changes the tilt on the times
		if (mouseClickY < y)
		{
			xRotation += dRotate;
		}

		if (mouseClickY > y)
		{
			xRotation -= dRotate;
		}
		
		// Update the display in realtime
		display();
	}
}

// Rotates the display in a timed animation
void animate(int degrees)
{    
	yRotation += degrees;
	display();

	// Don't go too fast so events can happen
    glutTimerFunc(50, animate, 5);
}

// Draws the planes that the maze fills in
void drawPlane(double z)
{
   glEnable(GL_DEPTH_TEST);

   int tilesize = 1;
   int xSize = 20;
   int ySize = 20;
   
   //double z = -20;

   glColor3f(1.0, 0.0, 0.0);

   int i = -1;

   glBegin(GL_QUADS);
   
   for (int x=-10; x<xSize-10; x++)
   {
       for (int y=-10; y<ySize-10; y++)
       {
          double r = tileArray[(++i)]; 
          double b = tileArray[(++i)];
          double g = tileArray[(++i)];

          if ((r == 1) || (g == 1) || (b == 1))
		  {
		  	  glColor3f(.5, .5, .5);	
          
   		      glVertex3d(x,y,z);
   		      glVertex3d(x+tilesize, y, z);
   		      glVertex3d(x+tilesize, y-tilesize, z);
   		      glVertex3d(x, y-tilesize, z); 

   		  }
       }
   }

   glEnd();
}

// Draws a plane (heh) old blue panel
void drawWall1(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4)
{
	glColor3f(0.0, 0.0, 1.0);

    glBegin(GL_QUADS);

	glVertex3d(cx1, cy1, cz1);
	glVertex3d(cx2, cy2, cz2);
	glVertex3d(cx3, cy3, cz3);
	glVertex3d(cx4, cy4, cz4);

    glEnd();
}

// Draws two triangles to make a panel, blue and yellow
void drawWall2(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4)
{

    glColor3f(0,0,1);
    
    glBegin(GL_TRIANGLES);
	glVertex3d(cx1, cy1, cz1);
	glVertex3d(cx2, cy2, cz2);
	glVertex3d(cx3, cy3, cz3);
	glEnd();

    glColor3f(1,1,0);
    
    glBegin(GL_TRIANGLES);
	glVertex3d(cx1, cy1, cz1);
	glVertex3d(cx3, cy3, cz3);	
	glVertex3d(cx4, cy4, cz4);
	glEnd();
}

// Draws a panel with a line through the middle
void drawWall3(	double cx1, double cy1, double cz1,
				double cx2, double cy2, double cz2,
				double cx3, double cy3, double cz3,
				double cx4, double cy4, double cz4)
{
    glColor3f(1,0,0);

	glBegin(GL_QUADS);
	glVertex3d(cx1, cy1, cz1);
	glVertex3d(cx2, cy2, cz2);
	glVertex3d(cx3, cy3, cz3);
	glVertex3d(cx4, cy4, cz4);
	glEnd();

    glColor3f(1,1,1);

	glLineWidth(2.0);

	glBegin(GL_LINES);
	glVertex3d(cx1, cy1, cz1);
	glVertex3d(cx3, cy3, cz3);
	glEnd();

}

// Snags mouse clicks for the mouse move callback
void mouse(int button, int state, int mouseX, int mouseY)
{
	// Snag the origin point for the control drag
    if (button == GLUT_LEFT_BUTTON)
	{
   		if (state == GLUT_DOWN)
   		{
			if (!leftMouseDown)
			{
				mouseClickX = mouseX;
	   			mouseClickY = mouseY;
			}
   		
   		    leftMouseDown = 1;
   		}
		
		if (state == GLUT_UP)
		{
			leftMouseDown = 0;
		}
   }
}

int main(int argc, char** argv)
{
   loadMaze();

   // Glut init
   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutInitWindowSize (500, 500); 
   glutInitWindowPosition (100, 100);
   glutCreateWindow ("MazeGen 3D");

   init ();

   // Callbacks
   glutDisplayFunc(display); 
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutTimerFunc(50, animate, 5);   
   glutMouseFunc(mouse);
   glutMotionFunc(mouseMove);
   glutMainLoop();
   
   return 0;
}




